#' ReadDataFile
#'
#' Read content of a data file, either Excel or text.\cr
#' It is a wrapper around \code{\link{ReadExcelFile}} and \code{\link{ReadTextFile}}. Appropriate
#' loading function is resolved by the extension of \code{fileName}.
#'
#' @param fileName Name of file to read. Required.
#' @param fileType Type of data to read. Required if \code{fileName} has no file extension,
#'   otherwise optional.
#' @param ... Additional parameters passed to \code{\link{ReadExcelFile}} or
#'   \code{\link{ReadTextFile}}. Optional.
#'
#' @return data.table object.
#'
#' @examples
#' \dontrun{
#' ReadDataFile(fileName)
#' ReadDataFile(fileName, "xlsx")
#' ReadDataFile(fileName, "csv")
#' }
#'
#' @export
ReadDataFile <- function(fileName, fileType, ...)
{
  stopifnot(!missing(fileName))

  # Get file type
  if (missing(fileType)) {
    fileType <- tolower(tools::file_ext(fileName))
  }

  if (fileType == "") {
    stop("Undetected file type")
  }

  # Run appropriate reading function
  data <- switch(fileType,
                 "xls"  = ReadExcelFile(fileName, fileType, ...),
                 "xlsx" = ReadExcelFile(fileName, fileType, ...),
                 "txt"  = ReadTextFile(fileName, ...),
                 "csv"  = ReadTextFile(fileName, ...),
                 {"Unsupported file extension"})

  return(data)
}
