list(
  # Adjustment name ----
  Name = "Reporting Delays",

  # Input parameters to the adjustment function ----
  Parameters = list(
    # Parameter 1: a specification with label, type and default value
    startYear = list(
      label = "Start year",
      value = 2000L,
      input = "numeric")
  ),

  # Names of packages that must be made available to the adjustment function ----
  RequiredPackageNames = c(),

  ## Adjustment function ----
  AdjustmentFunction = function(inputData, parameters) {

    startYear <- parameters$startYear
    endQrt <- 2017.25;

    outputData <- copy(inputData)

    outputData[, ":="(
      NotificationTime = DateOfNotificationYear + 1/4 * DateOfNotificationQuarter,
      DiagnosisTime = DateOfDiagnosisYear + 1/4 * DateOfDiagnosisQuarter
    )]
    outputData[, VarX := 4 * (NotificationTime - DiagnosisTime)]
    outputData[VarX < 0, VarX := NA]

    outputData[, ":="(
      MinNotificationTime = min(NotificationTime),
      MaxNotificationTime = max(NotificationTime)
    ), by = .(ReportingCountry)]

    outputData <- outputData[DiagnosisTime > max(startYear, MinNotificationTime) &
                               NotificationTime <= endQrt]

    outputData[, ":="(
      VarT = 4 * (MaxNotificationTime - DiagnosisTime),
      ReportingDelay = 1L,
      Tf = 4 * (MaxNotificationTime - max(startYear, MinNotificationTime))
    )]
    outputData[, ":="(
      VarXs = Tf - VarX,
      VarTs = Tf - VarT
    )]

    model <- outputData[, survival::Surv(time = VarTs,
                                         time2 = VarXs,
                                         event = ReportingDelay)]
    fit <- survival::survfit(model ~ 1)

    final <- data.table(Delay = fit$time,
                        P = fit$surv,
                        StErr = fit$std.err,
                        Var = fit$std.err^2)
    final[, Delay := max(Delay) - Delay - 1]
    final <- final[Delay >= 0]

    agregat <- outputData[, .(Count = .N),
                          by = .(DiagnosisTime, DateOfDiagnosisYear, ReportingCountry, VarT)]
    agregat <- merge(agregat, final,
                     by.x = c("VarT"),
                     by.y = c("Delay"))
    agregat[, ":="(
      EstCount = Count / P,
      EstCountVar = (Count * (Count + 1) / P^4 * Var) + Count * (1 - P) / P^2
    )]

    agregat[, ":="(
      LowerEstCount = EstCount - 1.96 * sqrt(EstCountVar),
      UpperEstCount = EstCount + 1.96 * sqrt(EstCountVar)
    )]

    agregatByYear <- agregat[, .(Count = sum(Count),
                                 EstCount = sum(EstCount),
                                 EstCountVar = sum(EstCountVar)),
                             by = .(YearOfDiagnosis = DateOfDiagnosisYear)]

    agregatByYear[, ":="(
      LowerEstCount = EstCount - 1.96 * sqrt(EstCountVar),
      UpperEstCount = EstCount + 1.96 * sqrt(EstCountVar)
    )]

    outPlot <-
      ggplot(agregat, aes(x = DiagnosisTime)) +
      geom_ribbon(aes(ymin = pmin(pmax(LowerEstCount, 0), max(EstCount)),
                      ymax = pmin(pmax(UpperEstCount, 0), max(EstCount)),
                      fill = "95% confidence interval\nfor estimated total count"),
                  alpha = 0.4) +
      geom_line(aes(y = Count, color = "Reported count"), size = 0.2) +
      geom_line(aes(y = EstCount, color = "Estimated total count"), size = 0.5) +
      scale_fill_manual("", values = "grey70") +
      scale_colour_manual("", values = c("grey50", "blue")) +
      ggtitle("Reported and estimated total count of cases") +
      xlab("Diagnosis date") +
      ylab("Count of HIV cases")

    outPlotYr <-
      ggplot(agregatByYear, aes(x = YearOfDiagnosis)) +
      geom_ribbon(aes(ymin = pmin(pmax(LowerEstCount, 0), max(EstCount)),
                      ymax = pmin(pmax(UpperEstCount, 0), max(EstCount)),
                      fill = "95% confidence interval\nfor estimated total count"),
                  alpha = 0.4) +
      geom_line(aes(y = Count, color = "Reported count"), size = 0.2) +
      geom_line(aes(y = EstCount, color = "Estimated total count"), size = 0.5) +
      scale_fill_manual("", values = "grey70") +
      scale_colour_manual("", values = c("grey50", "blue")) +
      ggtitle("Reported and estimated total count of cases") +
      xlab("Diagnosis year") +
      ylab("Count of HIV cases")

    artifacts <- list(OutputPlot = outPlot,
                      OutputPlotYr = outPlotYr)

    return(list(Table = agregat,
                Artifacts = artifacts))
  }
)
