# User interface
inputDataUploadUI <- function(id)
{
  ns <- NS(id)

  tagList(
    div(style = "margin-top: 15px"),
    box(
      width = 12,
      title = "Input data",
      solidHeader = FALSE,
      status = "primary",
      collapsible = TRUE,
      fluidRow(
        column(4, fileInput(ns("fileInput"),
                            width = "100%",
                            label = "File input:")),
        column(8, withSpinner(uiOutput(ns("origDataDetails")),
                              type = 7,
                              proxy.height = "50px")))),
    uiOutput(ns("attrMapping"))
  )
}

# Server logic
inputDataUpload <- function(input, output, session)
{
  # Get namespace
  ns <- session$ns

  vals <- reactiveValues(attrMapping = list(),
                         attrMappingStatus = list(),
                         attrMappingValid = FALSE,
                         origDataAttrs = c(),
                         inputData = NULL)

  # EVENT: Input data file name changed
  originalData <- eventReactive(input[["fileInput"]], {
    # Get reference to file input
    fileInput <- input$fileInput

    # Validate
    validate(need(fileInput, message = FALSE))

    # Get input data file name and type
    fileType <- tolower(tools::file_ext(fileInput$name))
    fileName <- fileInput$datapath

    # Read input data
    originalData <- ReadDataFile(fileName, fileType)

    vals$origDataAttrs <- colnames(originalData)
    vals$attrMapping <- GetPreliminaryAttributesMapping(originalData)

    return(originalData)
  }, ignoreNULL = TRUE)

  # Output original data details when they have changed
  output[["origDataDetails"]] <- renderUI({
    # Respond to originalData change
    originalData <- originalData()

    isolate({
      # Get reference to file input
      fileInput <- input$fileInput

      # Get input data details
      inputDataDetails <-
        fluidRow(
          column(3,
                 p(strong("Name:"), fileInput$name),
                 p(strong("Size:"), FormatObjectSize(fileInput$size)),
                 p(strong("Type:"), fileInput$type),
                 p(strong("Number of records:"), nrow(originalData))),
          column(9,
                 p(strong("Column names:")),
                 p(id = ns("colNamesP"),
                   paste(colnames(originalData), collapse = ", ")))
        )

      return(inputDataDetails)
    })
  })

  # Generate attributes mapping
  output[["attrMapping"]] <- renderUI({
    # Respond to originalData change
    originalData()

    isolate({

      ns <- session$ns

      shinyjs::hide("attrMappingInitText")

      mappingStatusMsg <- uiOutput(ns("attrMappingStatus"))
      descrText <- p("Please, provide mapping between attributes used internally by the tool (column \"Tool\") and data dimensions (column \"Data\").")

      attributesMappingRowsUI <- lapply(names(vals$attrMapping), function(analysisAttr) {
        selectOptions <- list(placeholder = paste("Select", analysisAttr, "attribute"))
        if (is.null(vals$attrMapping[[analysisAttr]])) {
          selectOptions <- list(
            placeholder = paste("Select", analysisAttr, "attribute"),
            onInitialize = I('function() { this.setValue(""); }'))
          validMsg <- "Not ok"
        }

        selectId <- paste0("attr_", analysisAttr)

        widget <- tags$tr(
          # Selection attribute
          tags$td(analysisAttr),
          # Selection data dimension
          tags$td(selectizeInput(ns(selectId),
                                 label = NULL,
                                 choices = vals$origDataAttrs,
                                 selected = vals$attrMapping[[analysisAttr]],
                                 options = selectOptions))
        )

        # EVENT: Selected item changed
        observeEvent(input[[selectId]], {
          vals$attrMapping[[analysisAttr]] <- input[[selectId]]
        })

        return(widget)
      })

      attributesMappingTableUI <- tags$table(
        id = ns("attrMappingTable"),
        tags$thead(
          tags$tr(
            tags$th(class = ns("attrNameCol"),
                    "Tool"),
            tags$th("Data")
          )
        ),
        tags$tbody(
          attributesMappingRowsUI
        )
      )

      # Wrap in a box
      attributesMappingUI <- box(
        width = 12,
        title = "Attributes mapping",
        solidHeader = FALSE,
        status = "warning",
        collapsible = TRUE,
        descrText,
        fluidRow(
          column(6, attributesMappingTableUI),
          column(6, mappingStatusMsg)
        ))

      return(attributesMappingUI)
    })
  })

  observeEvent(vals$attrMapping, {
    attrMappingStatus <- GetAttrMappingStatus(vals$attrMapping)
    vals$attrMappingValid <- attrMappingStatus$Valid
    vals$attrMappingStatus <- attrMappingStatus[c("NonMapped", "MultipleMapped")]
  })

  output[["attrMappingStatus"]] <- renderUI({
    nonMappedAnalysisAttrs <- vals$attrMappingStatus$NonMapped
    multipleMappedDataAttrs <- vals$attrMappingStatus$MultipleMapped
    attrMappingValid <- vals$attrMappingValid

    isolate({

      header <- NULL
      textColor <- NULL
      if (attrMappingValid) {
        textColor <- "seagreen"
        header <- p(tags$i(class = "material-icons", "check"), "Mapping is valid.", style = paste("color:", textColor))
      } else {
        textColor <- "red"
        header <- p(tags$i(class = "material-icons", "error_outline"), "Mapping is not valid. Please, correct the following issues:", style = paste("color:", textColor))
      }

      nonMappedStatus <- NULL
      if (length(nonMappedAnalysisAttrs) > 0) {
        nonMappedStatus <- tagList(
          tags$span("Tool attributes not mapped:", style = "font-weight: bold"),
          tags$ol(
            lapply(nonMappedAnalysisAttrs, function(analysisAttr) tags$li(analysisAttr))
          )
        )
      }

      multiMappedStatus <- NULL
      if (length(multipleMappedDataAttrs) > 0) {
        multiMappedStatus <- tagList(
          tags$span("Data attributes mapped multiple times:", style = "font-weight: bold"),
          tags$ol(
            lapply(names(multipleMappedDataAttrs), function(dataAttr) {
              analysisAttrs <- multipleMappedDataAttrs[[dataAttr]]
              tags$li(paste(AddQuoteMarks(dataAttr),
                            "mapped to",
                            paste(AddQuoteMarks(analysisAttrs), collapse = ", ")))
            })
          )
        )
      }

      attrMappingStatus <- box(width = 12,
                               style = "background-color: ghostwhite;",
                               header,
                               nonMappedStatus,
                               multiMappedStatus)

      return(attrMappingStatus)
    })
  })

  inputData <- eventReactive(vals$attrMapping, {
    intermediateData <- ApplyAttributesMapping(originalData(), vals$attrMapping)
    PreProcessInputData(intermediateData)
  })

  return(list(originalData = originalData,
              inputData = inputData))
}
